MODULE Modules;  (* jt 27.2.95 *)

  IMPORT SYSTEM, Console, Unix(*for errno*), Kernel;

  CONST
    ModNameLen* = 20;

  TYPE
    ModuleName* = ARRAY ModNameLen OF CHAR;
    Module* = POINTER TO ModuleDesc;
    Cmd* = POINTER TO CmdDesc;
    ModuleDesc* = RECORD  (* cf. SYSTEM.Mod *)
      next-: Module;
      name-: ModuleName;
      refcnt-: LONGINT;
      cmds-: Cmd;
      types-: LONGINT;
      enumPtrs-: PROCEDURE (P: PROCEDURE(p: LONGINT));
      reserved1, reserved2: LONGINT;
    END ;

    Command* = PROCEDURE;

    CmdDesc* = RECORD
      next-: Cmd;
      name-: ARRAY 24 OF CHAR;
      cmd-: Command
    END ;

  VAR
    res*: INTEGER;
    resMsg*: ARRAY 256 OF CHAR;
    imported*, importing*: ModuleName;


  PROCEDURE -include()
    "#include <dlfcn.h>";

  PROCEDURE -dlopen(path: ARRAY OF CHAR): LONGINT
    "(long)dlopen((const char*)path, RTLD_LAZY | RTLD_GLOBAL)";

  PROCEDURE -dlsym(handle: LONGINT; name: ARRAY OF CHAR): Command
    "(Modules_Command)dlsym((void*)handle, (const char*)name)";

  PROCEDURE -dlclose(handle: LONGINT): LONGINT
    "dlclose((void*)handle)";

  PROCEDURE -dlerror(VAR s: ARRAY OF CHAR)
    "__COPY(dlerror(), s, s__len)";

  PROCEDURE -next(): LONGINT
    "(LONGINT)RTLD_NEXT";

  PROCEDURE -modules*(): Module
      "(Modules_Module)SYSTEM_modules";

  PROCEDURE -setmodules*(m: Module)
      "SYSTEM_modules = m";


  PROCEDURE Append(VAR a: ARRAY OF CHAR; b: ARRAY OF CHAR);
    VAR i, j: INTEGER;
  BEGIN
    i := 0; WHILE a[i] # 0X DO INC(i) END;
    j := 0; WHILE b[j] # 0X DO a[i] := b[j]; INC(i); INC(j) END;
    a[i] := 0X
  END Append;

  PROCEDURE GetSubsys1(n: ARRAY OF CHAR; VAR s: ARRAY OF CHAR);  (* digit treated as upper case *)
    VAR i: INTEGER; ch: CHAR;
  BEGIN
    ch := n[0]; i := 0;
    WHILE (ch # 0X) & ((ch < "a") OR (ch > "z")) DO s[i] := ch; INC(i); ch := n[i] END ;
    WHILE (ch >= "a") & (ch <= "z") DO s[i] := ch; INC(i); ch := n[i] END ;
    IF ch = 0X THEN s[0] := 0X ELSE s[i] := 0X END
  END GetSubsys1;

  PROCEDURE GetSubsys2(n: ARRAY OF CHAR; VAR s: ARRAY OF CHAR);  (* digit treated as lower case *)
    VAR i: INTEGER; ch: CHAR;
  BEGIN
    ch := n[0]; i := 0;
    WHILE (ch >= "A") & (ch <= "Z") DO s[i] := ch; INC(i); ch := n[i] END ;
    WHILE (ch # 0X) & ((ch < "A") OR (ch > "Z")) DO s[i] := ch; INC(i); ch := n[i] END ;
    IF ch = 0X THEN s[0] := 0X ELSE s[i] := 0X END
  END GetSubsys2;

  PROCEDURE FullName(VAR n: ARRAY OF CHAR): BOOLEAN;
    VAR i: INTEGER; ch: CHAR;
  BEGIN
    i := 0; ch := n[0];
    WHILE ch # 0X DO
      IF ch = "." THEN RETURN TRUE END ;
      INC(i); ch := n[i]
    END ;
    RETURN FALSE
  END FullName;

  PROCEDURE err();
    VAR i: INTEGER; s: ARRAY 256 OF CHAR;
  BEGIN i := 0; dlerror(s);
    WHILE s[i] # 0X DO INC(i) END;
    IF (s[i-2] # "=") OR (s[i-1] # "2") THEN Console.String(s); Console.Ln END (* ignore "file not found" errors *)
  END err;

  PROCEDURE ThisMod* (name: ARRAY OF CHAR): Module;
    VAR m: Module; bodyname, libname, libname2: ARRAY 64 OF CHAR;
      body: Command; lib, handle: LONGINT;
  BEGIN m := modules(); handle := 0;
    WHILE (m # NIL) & (m.name # name) DO m := m.next END ;
    IF m = NIL THEN
      COPY(name, bodyname); Append(bodyname, "__init");
      body := dlsym(next(), bodyname); (* is it contained in one of the previously loaded shared objects? *)
      IF body = NIL THEN
        libname := "lib"; Append(libname, name); Append(libname, ".so");
        lib := dlopen(libname);
        IF lib # 0 THEN body := dlsym(lib, bodyname) ELSE err() END ;
        IF body # NIL THEN handle := lib END
      END ;
      IF body = NIL THEN
        GetSubsys1(name, libname);
        IF libname[0] # 0X THEN
          libname := "lib"; Append(libname, libname); Append(libname, ".so");
          lib := dlopen(libname);
          IF lib # 0 THEN body := dlsym(lib, bodyname) ELSE err() END
        END
      END ;
      IF body = NIL THEN
        GetSubsys2(name, libname2);
        IF libname2[0] # 0X THEN
          libname2 := "lib"; Append(libname2, libname2); Append(libname2, ".so");
          IF (libname2 # libname) THEN
            lib := dlopen(libname2);
            IF lib # 0 THEN body := dlsym(lib, bodyname) ELSE err() END
          END
        END
      END ;
      IF body = NIL THEN
        IF FullName(Kernel.LIB) THEN COPY(Kernel.LIB, libname)
        ELSE libname := "lib"; Append(libname, Kernel.LIB); Append(libname, ".so")
        END ;
        lib := dlopen(libname);
        IF lib # 0 THEN body := dlsym(lib, bodyname) ELSE err() END
      END ;
      IF body # NIL THEN
        body(); m := modules();
        WHILE (m # NIL) & (m.name # name) DO m := m.next END ;
        IF m # NIL THEN m.reserved1 := handle END
      END
    END ;
    IF m # NIL THEN res := 0; resMsg := ""
    ELSE res := 1; COPY(name, importing);
      resMsg := ' module "'; Append(resMsg, name); Append(resMsg, '" not found');
    END ;
    RETURN m
  END ThisMod;

  PROCEDURE ThisCommand* (mod: Module; name: ARRAY OF CHAR): Command;
    VAR c: Cmd;
  BEGIN c := mod.cmds;
    WHILE (c # NIL) & (c.name # name) DO c := c.next END ;
    IF c # NIL THEN res := 0; resMsg := ""; RETURN c.cmd
    ELSE res := 2; resMsg := ' command "'; COPY(name, importing);
      Append(resMsg, mod.name); Append(resMsg, "."); Append(resMsg, name); Append(resMsg, '" not found');
      RETURN NIL
    END
  END ThisCommand;

  PROCEDURE Free*(name: ARRAY OF CHAR; all: BOOLEAN);
    VAR m, p: Module;
  BEGIN m := modules();
    IF all THEN
      res := 1; resMsg := 'unloading "all" not yet supported'
    ELSE
      WHILE (m # NIL) & (m.name # name) DO p := m; m := m.next END ;
      IF (m # NIL) & (m.refcnt = 0) THEN
        IF m.reserved1 # 0 THEN
          IF dlclose(m.reserved1) # 0 THEN
            res := 1; dlerror(resMsg);
          ELSE res := 0;
            IF m = modules() THEN setmodules(m.next)
            ELSE p.next := m.next
            END
          END
        ELSE res := 1;
          resMsg := "module not loaded in separate library"
        END
      ELSE res := 1;
        IF m = NIL THEN resMsg := "module not found"
        ELSE resMsg := "clients of this module exist"
        END
      END
    END
  END Free;

END Modules.
